/* ----------------------------------------------------------------
 * routines for ecclesiastical calendar calculations 
 * used header jd.h
 * 
 * ----------------------------------------------------------------
 */


/*----------------------------------------------------------------
 * easter() checks to see which calendar we are on,
 * and calls the appropriate routine accordingly
 * ----------------------------------------------------------------
 */

void easter( year, month, day)
{
  dateSetGregorian( year, 4, 1 ); 
  if (gregorian)
    gregorianEaster( year, month, day );
  else
    julianEaster( year, month, day ); 

}



/* ----------------------------------------------------------------
 * gregorianEaster sets the date of easter; Expl. Supplement, p. 581-582
 * 
 * ----------------------------------------------------------------
 */
void
gregorianEaster( year, month, day )
int year;
int *month;
int *day;
{
  int c, n, k, i, j, el;

  c = year / 100;
  n = year - 19*(year/19);
  k = (c-17)/25;
  i = c - c/4 - (c-k)/3 + 19*n + 15;
  i = i - 30*(i/30);
  i = i - (i/28) * (1 - (i/28) * (29/(i+1)) * ((21-n)/11));
  j = year + year/4 + i + 2 - c + c/4;
  j = j - 7*(j/7);
  el = i - j;

  *month = 3 + (el+40)/44;
  *day   = el + 28 - 31*(*month/4);

}

long
jdEaster( year )
int year;
{
  int month, day;
  long jd; 

  easter( year, &month, &day );
  if (gregorian)
    jd = GregorianCalendarToJD( year, month, day ); 
  else
    jd = JulianCalendarToJD( year, month, day ); 

  return jd;

}



/* ----------------------------------------------------------------
 * julianEaster is more complicated . . .
 * source: _Explanatory Supplement_ (1961), pp. 422-424
 * the same method is provided in the 1928 _Book of Common Prayer_,
 *    but without the table to go from Golden Number to date of paschal new moon
 *    for the julian calendar.
 * ----------------------------------------------------------------
 */

void
julianEaster( year, month, day )
int year;
int *month;
int *day;
{
  int gn;   /* golden number */
  int moonMonth;    /* month of the paschal full moon */
  int moonDay;	    /* day of the paschal full moon */
  static firsttime=1;		/* flag that tells if this is first time through this subroutine */ 
  static int paschalNMMonth[20];	  /* month/day of the paschal new moon, as function */
  static int paschalNMDay[20];		  /* of the golden number */ 
  long jdNewMoon, jdFullMoon, jdEaster;	  /* jd of paschal new moon, full moon, easter */ 
  int offset; 

  if (firsttime)
  {
    /* the applicable parts of table 14.4, p. 422, _Expl Supp_ (1961): */ 
    paschalNMMonth[1] = 3;
    paschalNMDay[1]   = 23; 

    paschalNMMonth[2] = 3;
    paschalNMDay[2]   = 12; 

    paschalNMMonth[3] = 3;
    paschalNMDay[3]   = 31; 

    paschalNMMonth[4] = 3;
    paschalNMDay[4]   = 20; 

    paschalNMMonth[5] = 3;
    paschalNMDay[5]   = 9; 

    paschalNMMonth[6] = 3;
    paschalNMDay[6]   = 28; 

    paschalNMMonth[7] = 3;
    paschalNMDay[7]   = 17; 

    paschalNMMonth[8] = 4;
    paschalNMDay[8]   = 5; 

    paschalNMMonth[9] = 3;
    paschalNMDay[9]   = 25; 

    paschalNMMonth[10] = 3;
    paschalNMDay[10]   = 14; 

    paschalNMMonth[11] = 4;
    paschalNMDay[11]   = 2; 

    paschalNMMonth[12] = 3;
    paschalNMDay[12]   = 22; 

    paschalNMMonth[13] = 3;
    paschalNMDay[13]   = 11; 

    paschalNMMonth[14] = 3;
    paschalNMDay[14]   = 30; 

    paschalNMMonth[15] = 3;
    paschalNMDay[15]   = 19;
    
    paschalNMMonth[16] = 3;
    paschalNMDay[16]   = 8; 

    paschalNMMonth[17] = 3;
    paschalNMDay[17]   = 27; 

    paschalNMMonth[18] = 3;
    paschalNMDay[18]   = 16; 

    paschalNMMonth[19] = 4;
    paschalNMDay[19]   = 4; 

    firsttime=0; 
  }

  gn = goldenNumber( year );
  moonMonth = paschalNMMonth[ gn ];
  moonDay   = paschalNMDay[ gn ];

  jdNewMoon  = JulianCalendarToJD( year, moonMonth, moonDay );
  jdFullMoon  = jdNewMoon + 13;

  offset = 7 - weekdayNumber(jdFullMoon);
  jdEaster   = jdFullMoon + offset; 

  JDtoJulianCalendar( jdEaster, &year, month, day ); 

}



int
goldenNumber( year )
int year;
{
  int gn; 
  gn = ((year + 1) % 19 );
  if (gn == 0) gn = 19;
  return gn; 
}



/* ----------------------------------------------------------------
 * these should be simple jumps from easter . . .
 * ----------------------------------------------------------------
 */

char *
ashwednesday( year )
int year;
{
  static char buf[20];
  return fromEaster( year, buf, -46 ); 
}


char *
pentecost( year )
int year;
{
  static char buf[20];
  return fromEaster( year, buf, 49 ); 
}

char *
ascension( year )
int year;
{
  static char buf[20];
  return fromEaster( year, buf, 39 ); 
}

char *
trinity( year )
int year;
{
  int offset;
  static char buf[20];
  return fromEaster( year, buf, 56 ); 
}

char *
septuagesima( year )
int year;
{
  int offset;
  static char buf[20];
  return fromEaster( year, buf, -63 ); 
}

char *
sexagesima( year )
int year;
{
  int offset;
  static char buf[20];
  return fromEaster( year, buf, -56 ); 
}

char *
quinquagesima( year )
int year;
{
  int offset;
  static char buf[20];
  return fromEaster( year, buf, -49 ); 
}



char *
fromEaster( year, buf, offset )
int year;
char *buf; 
int offset;
{
  long jd; 
  int month, day;
  char *format; 
  int y; 
  
  jd = jdFromEaster( year, offset ); 
  JDtoCalendar( jd, &y, &month, &day );
/*  printf("\n month = %d  \n", month );
  printf(" day   = %d  \n", day ); 
*/
  /* construct the return string: */ 
  sprintf( buf, "%d-%d", month, day );

  return buf; 

}

long
jdFromEaster( year, offset )
int year;
int offset;
{
  long jd; 
  int month, day;

  easter( year, &month, &day ); 
  jd = CalendarToJD( year, month, day );
  jd += offset;

  return jd; 
}

long 
jdAdvent1( year )
int year;
{
  long jdStAndrew; 
  int month, day;
  long jdAdvent; 
  int iweekday; 

  jdStAndrew = CalendarToJD( year, 11, 30 );
  iweekday = weekdayNumber( jdStAndrew );  

  if (iweekday <= 3)
    jdAdvent = jdStAndrew - iweekday;
  else
    jdAdvent = jdStAndrew - iweekday + 7; 

  return jdAdvent; 

}

char *
advent1string( year )
int year;
{
  long jd;
  double djd;
  static char buf[12]; 
  int xyear, month, day;

  jd = jdAdvent1( year );     /* try the Gregorian calendar */ 
  setGregorian( jd );
  jd = jdAdvent1( year );     /* in case we are on the Julian calendar */ 

  JDtoCalendar( jd, &xyear, &month, &day );
  sprintf( buf, "%d-% d-% d", xyear, month, day ); 

  return buf; 

}


long
jdXmas( year )
int year;
{
  long xmas; 

  xmas = CalendarToJD( year, 12, 25 ); 
  return xmas;

}




int 
weektrin( jd )
long jd;
{
  long jdeaster;
  long jdtrinity;
  int year, month, day;
  long diffjd; 

  JDtoCalendar( jd, &year, &month, &day );
  easter( year, &month, &day );
  jdeaster = CalendarToJD( year, month, day );
  jdtrinity = jdeaster + 56;

  diffjd = jd - jdtrinity;
  diffjd = diffjd / 7;	    /* converts to weeks after trinity, and truncates */


  return (int) diffjd;

}


/* kweek should return a string designating the week of the liturgical year;
 * i. e., the week of the "kalendar"
 * (why is the ecclesiastical calendar spelled with a 'k' ??)
 */


char *
kweek( jd )
long jd;
{
  int weekno;
  static char buf[130]; 
  int year, month, day; 

  /* julian dates of various dates in the year: */ 
  long xmas, easter, advent, pentecost, trinity, ascension, ashwednesday; 
  long septuagesima, sexagesima, quinquagesima, epiphany, epiphany1, lent1; 
  long xmas1; 

  static char commonLectionary[32];	/* to hold name of sunday in Ordinary Time */ 
  int commonCount;			/* number of sunday in Ordinary Time */ 
  int proper;				/* Proper Number for the 1979 Book of Common Prayer  */ 
  long proper1;				/* date of Sunday of Proper 1 */
  long previousSunday;			/* date of sunday previous to jd */
  long ordinary9;			/* date of sunday 9 of ordinary time */ 

  /* this next variable is used in epiphany season, to get around what appears to be a bug
   * in the Borland C compiler ??!!
   * -- I've never run into a compiler-bug before.  it must have been mine.
   */
  int itemp;	/* a temp storage for an integer */ 


  JDtoCalendar( jd, &year, &month, &day ); 
  xmas	      = jdXmas( year ); 
  xmas1	      = xmas + (7 - weekdayNumber(xmas)); 
  advent      = jdAdvent1( year ); 

  easter      = jdEaster( year );
  trinity     = easter + 56;
  pentecost   = easter + 49;
  ascension   = easter + 39; 
  ashwednesday = easter - 46;
  lent1	      = ashwednesday + 4;

  epiphany    = CalendarToJD( year, 1, 6 ); 
  epiphany1   = epiphany + (7 - weekdayNumber( epiphany ));
  septuagesima = easter - 63;
  sexagesima  = easter - 56;
  quinquagesima = easter - 49;


  /* -------------------------------- */ 
  /* which Proper or Sunday of Ordinary Time: */ 

  if ( jd >= epiphany1 && jd < ashwednesday-3 )
  {
    commonCount = 1 + (jd-epiphany1)/7;
    sprintf( commonLectionary, "Ordinary Time %d", commonCount ); 
  }
  if ( jd >= trinity+7 && jd < advent )
  {
    /* the formula is inferred from _The Revised Common Lectionary_ (Abingdon, 1992), p. 14;
     * that's the best source (only source) I have . . .
     */ 
    ordinary9 = CalendarToJD( year, 6, 4 );
    ordinary9 = ordinary9 - weekdayNumber( ordinary9 );
    if ( jd >= ordinary9)
      commonCount = 9 + (jd - ordinary9)/7; 
    else
      commonCount = 8 - (ordinary9 - jd)/7; 

    previousSunday = jd - weekdayNumber( jd );	  /* jd of Sunday previous to input argument */ 
    /* proper 1 falls on the sunday nearest to May 11: */ 
    proper1 = CalendarToJD( year, 5, 11 );	  /* jd of Sunday of Proper 1 */ 
    if ( weekdayNumber( proper1 ) <= 3 )
      proper1 -= weekdayNumber( proper1 );
    else
      proper1 = proper1 - weekdayNumber( proper1 ) + 7; 
    proper = 1 + (previousSunday - proper1) / 7; 

    sprintf( commonLectionary, "Proper %d, Ordinary Time %d", proper, commonCount ); 
  }


  /* -------------------------------- */
  /* From new-year to Lent */
  if (jd < epiphany)
  {
    return "before epiphany"; 
  }
  if (jd == epiphany)
    return "Epiphany";
  if (jd > epiphany && jd < epiphany1)
    return "Epiphany"; 
  if (jd >= epiphany1 && jd < septuagesima)
  {
    itemp = 1+(jd - epiphany1)/7;
    sprintf( buf, "Epiphany %d, %s", itemp, commonLectionary );
    return buf;
  }
  if (jd >= septuagesima && jd < sexagesima)
  {
    itemp = 1+(jd - epiphany1)/7;
    sprintf( buf, "Epiphany %d, Septuagesima, %s", itemp, commonLectionary );
    return buf;
  }
  if (jd >= sexagesima && jd < quinquagesima)
  {
    itemp = 1+(jd - epiphany1)/7;
    sprintf( buf, "Epiphany %d, Sexagesima, %s", itemp, commonLectionary );
    return buf;
  }
  if (jd >= quinquagesima && jd < ashwednesday)
  {
    sprintf( buf, "Epiphany %d, Quinquagesima", 1+(jd - epiphany1)/7 );
    return buf;
  }



  /* -------------------------------- */
  /* Lent-Easter Season */ 
  if (jd < trinity && jd >= pentecost )
    return "Pentecost";
  if (jd == easter)
    return "Easter Sunday";

  if (jd < pentecost && jd > easter)
  {
    if (jd != ascension)
    {
      itemp = (jd - easter)/7; 
      sprintf( buf, "%d of / %d after Easter",   itemp+1, itemp ); 
    }
    else
      sprintf( buf, "Ascension" ); 
    return buf;
  }


  if (jd < easter-3 && jd >= easter - 6)
    return "Holy Week";
  if (jd == easter-1)
    return "Holy Saturday";
  if (jd == easter-2)
    return "Good Friday";
  if (jd == easter-3)
    return "Maundy Thursday";
  if (jd == easter-7)
    return "Palm Sunday"; 

  if (jd == ashwednesday)
    return "Ash Wednesday";
  if (jd > ashwednesday && jd < lent1)
    return "after Ash Wednesday"; 

  if (jd >= lent1 && jd < easter - 7)
  {
    sprintf( buf, "Lent % d", 1 + (jd - lent1)/7 );
    return buf;
  }


  /* -------------------------------- */
  /* after Trinity: */ 
  if (jd == trinity) 
    return "Trinity Sunday";
  if (jd > trinity && jd < trinity+7)
    return "Trinity";
  if (jd >= trinity+7 && jd < advent)
  {
     sprintf( buf, "Trinity %d, %s", weektrin( jd ), commonLectionary );
     return buf;
  }
 


  /* -------------------------------- */
  /* advent to year-end: */ 
 if (jd >= advent && jd < xmas)
  {
    /* advent weeks are numbered 1-4, not 0-3: */ 
    sprintf( buf, "Advent % d", 1 + (jd - advent)/7 );
    return buf;
  }
  if (jd == xmas)
  {
    sprintf( buf, "Christmas Day" ); 
    return buf;
  }
  if (jd > xmas && jd < xmas1)
    return "Christmas";
  if (jd >= xmas1)
    return "Christmas 1"; 


  printf(" you should not have gotten here in kweek() \n");
  return "error"; 
}


/* ---------------------------------------------------------------- */

void 
kdates( year )
int year;
{
  int month, day; 

  easter( year, &month, &day );
  printf(" Ash Wednesday   %s \n",  ashwednesday( year ) ); 
  printf(" Easter Day      %d-%d \n", month, day ); 
  printf(" Ascension Day   %s \n",  ascension( year )    );
  printf(" Pentecost       %s \n",  pentecost( year )    );
  printf(" Trinity Sunday  %s \n",  trinity( year )      );
  printf(" Advent One      %s \n",  advent1string( year )); 

}


/* ---------------------------------------------------------------- */

void
yeartable( year )
int year;
{
  long jd;
  double jdx;
  long jdstart;
  int month, day;

  jdstart = GregorianCalendarToJD( year, 1, 1 ); 
  setGregorian( jdstart );
  if (!gregorian)
    jdstart = JulianCalendarToJD( year, 1, 1 ); 

  printf("\n\n");
  for (jd=jdstart; jd<jdstart+366; ++jd)
  {
    setGregorian( jd ); 
    JDtoCalendar( jd, &year, &month, &day ); 
    printf(" jd %ld = %s % d % d % d ", jd, weekday(jd), year, month, day );
    printf(" %s \n", kweek( jd ) );
  }

}



/* returns the year of the 3-year eucharistic lectionary: 'A', 'B', 'C' */ 

int 
kyear( jd )
long jd;
{
  long advent1;	  /* jd of first sunday of advent */
  int year, month, day;
  int cycle; 

  setGregorian( jd ); 
  JDtoCalendar( jd, &year, &month, &day ); 
  advent1 = jdAdvent1( year );


  if (jd >= advent1)
    cycle = (year % 3);
  else
    cycle = ((year-1) % 3);

  cycle += 'A'; 

  return cycle;

}

